import os
import os.path
import urllib.parse

import batou.component
import batou.lib.file
import batou.lib.git

import batou_ext.ssh
from batou_ext.file import SymlinkAndCleanup


class GitCheckout(batou.component.Component):
    """
    Performing a git checkout via SSH and sync into a special folder to
    add customized files later on.

    Usage:
    The component is depending on SSH so it expects a SSH keypair you
    could provide via batou_ext.ssh.SSHKeyPair.

    Add GitCheckout-component to your custom component

    self.checkout = GitCheckout(
        git_clone_url='ssh://git@git.example.org:12345/example-repo.git',
        git_revision='1234567892134131313132231',
        git_target=self.map('checkout'),
        git_port=12345,
        exclude=('src/my.conf', '.gitingore', 'db/dump.sql'))
    self += self.checkout

    This will create a folder 'checkout/' inside your component as well
    as a folder prepared-1234567892134131313132231.

    Now you can add your custom config:

    self += batou.lib.file.File(
        '{}/src/my.conf'.format(self.checkout.prepared_path),
        content='nice=true')

    The option sync_parent_folder is allowing you to sync into a different
    folder than the current one.
    """

    _required_params_ = {
        "git_clone_url": "https://example.com/repos",
        "git_revision": "commit-hash",
        "git_target": ".",
    }
    git_host = None
    git_clone_url = None
    git_revision = None
    git_target = None
    git_port = None
    exclude = ()
    sync_parent_folder = None

    # Automatically scan the remote ssh hostkey (once)
    scan_host = True

    def configure(self):

        # We need to ensure to have a valid URL. Assuming URL with schema
        # should be valid.
        assert urllib.parse.urlparse(self.git_clone_url).scheme

        if self.scan_host:
            if not self.git_host:
                self.git_host = urllib.parse.urlparse(
                    self.git_clone_url
                ).hostname
            if not self.git_port:
                self.git_port = (
                    urllib.parse.urlparse(self.git_clone_url).port or 22
                )
            # Add remote host to known hosts
            self += batou_ext.ssh.ScanHost(self.git_host, port=self.git_port)

        # Check whether we need a SSH key
        if urllib.parse.urlparse(self.git_clone_url).scheme == "ssh":
            self.require("sshkeypair")

        # Get a recent checkout
        self += batou.lib.git.Clone(
            self.git_clone_url,
            revision=self.git_revision,
            target=self.git_target,
        )

        # Actually sync into a working copy where parent-component can
        # add custom files
        if self.sync_parent_folder:
            self.prepared_path = self.map(
                "{}/prepared-{}".format(
                    self.sync_parent_folder, self.git_revision
                )
            )
        else:
            self.prepared_path = self.map(
                "prepared-{}".format(self.git_revision)
            )
        self += batou.lib.file.Directory(self.prepared_path, leading=True)
        self += batou.lib.file.Directory(
            self.prepared_path,
            source=self.git_target,
            exclude=((".git",) + self.exclude),
        )

    def symlink_and_cleanup(self):
        return SymlinkAndCleanup(self.prepared_path, pattern="prepared-*")


class Commit(batou.component.Component):
    """Commit a file."""

    _required_params_ = {"message": "text"}
    namevar = "filename"
    message = None
    workingdir = "."
    author_name = "Batou"
    author_email = "batou@flyingcircus.io"

    def configure(self):
        assert self.message

    def verify(self):
        if self.has_changes():
            raise batou.UpdateNeeded()

    def update(self):
        with self.chdir(self.workingdir):
            self.cmd("git config user.email '{{component.author_email}}'")
            self.cmd("git config user.name '{{component.author_name}}'")
            self.cmd("git add {{component.filename}}")
            self.cmd(
                "git commit -m '{{component.message}}' {{component.filename}}"
            )

    def has_changes(self):
        with self.chdir(self.workingdir):
            stdout, stderr = self.cmd(
                "git status --porcelain {{component.filename}}"
            )
        return bool(stdout.strip())


class Remote(batou.component.Component):
    """
    Ensure that a git repository checkout out at `git_repo` has the origin
    `name` pointing to `url`.

    Usage::

        self += batou_ext.git.Remote(
            "/path/to/linux",
            name="upstream",
            url="ssh://git@github.com/torvalds/linux"
        )

    By default, the component aborts with an error if `git_repo` does not
    exist or is not a git repository. This is not always desirable, for
    instance when this is involved in migrating an existing `git_repo`
    to a new remote: then, this works fine for migration, but not when
    bootstrapping e.g. a new environment where `git_repo` doesn't exist
    yet.

    To solve that, it's possible to let the component do nothing if
    `git_repo` doesn't exist like this:

        self += batou_ext.git.Remote(
            "/path/to/linux",
            ignore_not_existing=True,
            # ...
        )
    """

    namevar = "git_repo"
    url = batou.component.Attribute(str)
    name = batou.component.Attribute(str, "origin")
    ignore_not_existing = batou.component.Attribute(bool, False)

    def verify(self):
        if not os.path.exists(self.git_repo):
            if self.ignore_not_existing:
                return
            else:
                raise ValueError(
                    f"batou_ext.git.Remote({self.git_repo}): path does not exist!"
                )
        elif not os.path.exists(f"{self.git_repo}/.git"):
            raise ValueError(
                f"batou_ext.git.Remote({self.git_repo}): not a git repository!"
            )

        with self.chdir(self.git_repo):
            stdout, _ = self.cmd(
                f"git remote get-url {self.name}", ignore_returncode=True
            )
            if self.url != stdout.strip():
                raise batou.UpdateNeeded()

    def update(self):
        with self.chdir(self.git_repo):
            stdout, _ = self.cmd("git remote")
            verb = "set-url" if self.name in stdout.splitlines() else "add"
            self.cmd(f"git remote {verb} {self.name} {self.url}")


class Push(batou.component.Component):
    """`git push` if there are outgoing changes."""

    workingdir = "."

    def verify(self):
        if self.has_outgoing_changesets():
            raise batou.UpdateNeeded()

    def update(self):
        with self.chdir(self.workingdir):
            self.cmd("git push")

    def has_outgoing_changesets(self):
        with self.chdir(self.workingdir):
            stdout, stderr = self.cmd("LANG=C git status")
        return "Your branch is ahead of" in stdout


class StopDeployOnLocalGitChange(batou.component.Component):
    """
    Helping to perform a git checkout to a target, that is not clean.

    batou.lib.git.Clone() might overwrite local changes inside target directory -- whereas you want to stop the process instead.

    Usage:
    The component can be used before perfoming the actual git clone:

        from batou.lib.git impot Clone
        from batou_ext.git import StopDeployOnLocalGitChange

        …
        target = "/my/target/path"
        self += StopDeployOnLocalGitChange(target)
        self += Clone(
            self.git_url,
            revision="01234567abcd",
            target=target)
    """

    namevar = "target"
    target = None

    def verify(self):
        if not os.path.exists(self.target):
            return
        if not os.path.exists(os.path.join(self.target, ".git")):
            return
        with self.chdir(self.target):
            stdout, stderr = self.cmd("git status --porcelain")
        changes = bool(stdout.strip())
        if changes:
            raise RuntimeError(f"Deployment aborted:\n{stdout}\n{stderr}")



from textwrap import dedent
from unittest import mock

import batou_ext.jenkins


def test_set_versions_git_mode(tmpdir):
    ini = tmpdir / "versions.ini"
    ini.write_text(
        dedent(
            """
                       [prog1]
                       # defaults to git-resolve for bw compat
                       url = git://prog1

                       [prog2]
                       url = http://prog2
                       update = git-resolve
                   """
        ),
        encoding="UTF-8",
    )

    with mock.patch("batou_ext.jenkins.git_resolve") as git_resolve:
        git_resolve.return_value = "abcdef"
        batou_ext.jenkins.set_versions(
            str(ini), '{"prog1": "a-tag", "prog2": ""}'
        )

    assert (
        dedent(
            """\
        [prog1]
        url = git://prog1
        revision = abcdef
        version = a-tag

        [prog2]
        url = http://prog2
        update = git-resolve
    """
        )
        == ini.read_text(encoding="UTF-8")
    )


def test_simple_value_update_mode(tmpdir):
    ini = tmpdir / "versions.ini"
    ini.write_text(
        dedent(
            """
                          [prog1]
                          version = 1.0
                          update = pass:version

                          [prog2]
                          url = http://prog2
                          update = pass:url
            """
        ),
        encoding="UTF-8",
    )
    batou_ext.jenkins.set_versions(
        str(ini), '{"prog1": "1.2", "prog2": "http://foobar"}'
    )
    assert (
        dedent(
            """\
      [prog1]
      version = 1.2
      update = pass:version

      [prog2]
      url = http://foobar
      update = pass:url
    """
        )
        == ini.read_text(encoding="UTF-8")
    )



import os
import os.path

import batou
import batou.utils
from batou.component import Attribute, Component
from batou.lib.file import File, Purge


class SSHKeyPair(Component):
    """Install SSH user and host keys.

    User keys are read from the secrets file and written to
    ~/.ssh/id_rsa{,.pub} and/or ~/.ssh/id_ed25519{,.pub}."""

    # RSA-keys
    id_rsa = None
    id_rsa_pub = None

    # ed25510-keys
    id_ed25519 = None
    id_ed25519_pub = None

    scan_hosts = Attribute("list", default=[])
    provide_itself = Attribute("literal", default=True)
    purge_unmanaged_keys = Attribute("literal", default=False)

    def configure(self):
        if self.provide_itself:
            self.provide("sshkeypair", self)

        self += File("~/.ssh", ensure="directory", mode=0o700)

        # RSA
        if self.id_rsa:
            self += File(
                "~/.ssh/id_rsa",
                content=(self.id_rsa + "\n"),
                mode=0o600,
                sensitive_data=True,
            )
        elif self.purge_unmanaged_keys:
            self += Purge("~/.ssh/id_rsa")

        if self.id_rsa_pub:
            self += File("~/.ssh/id_rsa.pub", content=self.id_rsa_pub)

        # ED25519
        if self.id_ed25519:
            self += File(
                "~/.ssh/id_ed25519",
                content="{}\n".format(self.id_ed25519),
                mode=0o600,
                sensitive_data=True,
            )

        elif self.purge_unmanaged_keys:
            self += Purge("~/.ssh/id_ed25519")

        if self.id_ed25519_pub:
            self += File("~/.ssh/id_ed25519.pub", content=self.id_ed25519_pub)

        # ScanHost
        for host in self.scan_hosts:
            self += ScanHost(host)


class ScanHost(Component):
    """This component adds the host key of a provided host to the service
    user's known host file."""

    namevar = "hostname"
    known_hosts = "~/.ssh/known_hosts"
    port = 22

    def configure(self):
        self.known_hosts = self.map(self.known_hosts)
        self += File(
            os.path.dirname(self.known_hosts),
            ensure="directory",
            mode=0o700,
            leading=True,
        )

    def verify(self):
        if not os.path.exists(self.known_hosts):
            raise batou.UpdateNeeded()
        with open(self.known_hosts, "r") as f:
            content = f.read()
        if self.port == 22:
            match = self.hostname
        else:
            match = "[{}]:{}".format(self.hostname, self.port)
        if match not in content:
            raise batou.UpdateNeeded()

    def update(self):
        try:
            del os.environ["SSH_AUTH_SOCK"]
        except KeyError:
            pass

        v4_failed = v6_failed = None
        try:
            self.cmd(
                'ssh-keyscan -4 -p {} "{}" >> "{}"'.format(
                    self.port, self.hostname, self.known_hosts
                )
            )
        except batou.utils.CmdExecutionError as e_v4:
            v4_failed = e_v4
        try:
            self.cmd(
                'ssh-keyscan -6 -p {} "{}" >> "{}"'.format(
                    self.port, self.hostname, self.known_hosts
                )
            )
        except batou.utils.CmdExecutionError as e_v6:
            v6_failed = e_v6

        if v4_failed and v6_failed:
            raise RuntimeError(
                f"Could not scan host {self.hostname}\n"
                f"IPv4: {v4_failed}\n"
                f"IPv6: {v6_failed}\n",
                v4_failed,
                v6_failed,
            )



from batou.component import Component

from batou_ext.postgres import DB, User


class RoundcubeDatabase(Component):

    _required_params_ = {
        "password": "tiger",
    }
    username = "roundcube"
    password = None
    database = "roundcube"

    dbms = "pgsql"

    command_prefix = ""
    locale = "en_US.utf8"

    def configure(self):
        dbserver = self.require_one("postgres")
        self.address = dbserver.address
        self.provide("roundcube::database", self)
        self += User(self.username, password=self.password)
        self += DB(self.database, owner=self.username, locale=self.locale)



"""RabbitMQ management.

Example::

    class RabbitMQ(batou.component.Component):

        erlang_cookie = None
        vhost = None
        username = None
        password = None

        def configure(self):
            self.provide('rabbitmq', self)
            self.address = batou.utils.Address(self.host.fqdn, 5672)
            self += ErlangCookie(cookie=self.erlang_cookie)
            self += PurgeUser('guest')
            self += VHost(self.vhost)
            self += User(
                self.username, password=self.password, tags=['management'])
            self += Permissions(
                self.username, permissions={self.vhost: ('.*', '.*', '.*')})

"""

import os
import stat

import batou
import batou.component
import batou.lib.file


class ErlangCookie(batou.component.Component):

    path = "~/.erlang.cookie"
    cookie = None

    def configure(self):
        self.path = self.map(self.path)
        self += batou.lib.file.Presence(self.path)

    def verify(self):
        try:
            with open(self.path, "r") as target:
                current = target.read()
                if current != self.cookie:
                    raise batou.UpdateNeeded()
            current = os.stat(self.path).st_mode
            if stat.S_IMODE(current) != 0o400:
                raise batou.UpdateNeeded()
        except FileNotFoundError:
            raise batou.UpdateNeeded()

    def update(self):
        os.chmod(self.path, 0o600)
        with open(self.path, "w") as target:
            target.write(self.cookie)
        os.chmod(self.path, 0o400)


class VHost(batou.component.Component):

    namevar = "name"
    name = None

    def verify(self):
        stdout, stderr = self.cmd("rabbitmqctl -q list_vhosts")
        vhosts = stdout.splitlines()
        if self.name not in vhosts:
            raise batou.UpdateNeeded()

    def update(self):
        self.cmd("rabbitmqctl add_vhost {{component.name}}")


class Permissions(batou.component.Component):

    namevar = "username"
    username = None
    permissions = None

    def verify(self):
        stdout, stderr = self.cmd(
            "rabbitmqctl -q list_user_permissions {{component.username}}"
        )
        lines = stdout.splitlines()
        to_validate = self.permissions.copy()
        self.to_update = []
        self.to_delete = []
        for line in lines:
            if not line or line == "vhost\tconfigure\twrite\tread":
                # Table header. This used to be hidden with `-q` but was
                # moved to `-s` in recent rabbit mq versions. Alas old
                # versions don't support `-s`. So filter manually.
                continue
            vhost, conf, write, read = line.split("\t", 3)
            perms = to_validate.pop(vhost, None)
            if perms is None:
                self.to_delete.append(vhost)
            elif perms != (conf, write, read):
                self.to_update.append(vhost)
        self.to_update.extend(list(to_validate.keys()))
        if self.to_update or self.to_delete:
            raise batou.UpdateNeeded()

    def update(self):
        for vhost in self.to_delete:
            self.cmd(
                self.expand(
                    "rabbitmqctl clear_permissions"
                    " -p {{vhost}} {{component.username}}",
                    vhost=vhost,
                )
            )
        for vhost in self.to_update:
            conf, write, read = self.permissions[vhost]
            self.cmd(
                self.expand(
                    "rabbitmqctl set_permissions"
                    " -p {{vhost}} {{component.username}}"
                    " '{{conf}}' '{{write}}' '{{read}}'",
                    vhost=vhost,
                    conf=conf,
                    write=write,
                    read=read,
                )
            )


class User(batou.component.Component):
    """Create rabbitmq user."""

    _required_params_ = {
        "tags": (),
    }
    namevar = "username"
    username = None
    password = None
    tags = None

    def configure(self):
        self.tags = sorted(self.tags)
        self.tags_str = " ".join(self.tags)

    def verify(self):
        stdout, stderr = self.cmd("rabbitmqctl -q list_users")
        users = stdout.splitlines()
        self.create = True
        self.set_tags = True
        for line in users:
            if not line or line.startswith("user\t"):
                continue
            user, tags = line.split("\t", 1)
            tags = sorted(tags[1:-1].split(", "))
            if user == self.username:
                self.create = False
                if tags == self.tags:
                    self.set_tags = False
                break
        if self.create or self.set_tags:
            raise batou.UpdateNeeded()

    def update(self):
        if self.create:
            self.cmd(
                "rabbitmqctl add_user"
                " '{{component.username}}' '{{component.password}}'"
            )
        if self.set_tags:
            self.cmd(
                "rabbitmqctl set_user_tags"
                " {{component.username}} {{component.tags_str}}"
            )


class PurgeUser(batou.component.Component):

    namevar = "username"
    username = None

    def verify(self):
        stdout, stderr = self.cmd("rabbitmqctl -q list_users")
        users = stdout.splitlines()
        for line in users:
            if not line or line.startswith("user\t"):
                continue
            user, tags = line.split("\t", 1)
            if user == self.username:
                raise batou.UpdateNeeded()

    def update(self):
        self.cmd("rabbitmqctl delete_user {{component.username}}")



from pathlib import Path

from batou.component import Component
from batou.lib.file import Purge

from batou_ext.nix import NixFile, component_to_nix, nix_dict_to_nix

# XXX: error messages stemming from the "batouModule" are displayed with
# wrong location information, it shows the glue module instead of the actual
# module. Can we fix that somehow?
GLUE_MODULE_TEMPLATE = """\
{{ pkgs, ... }}@args:

with builtins;

let
  moduleArgs = (import {workdir}/{prefix}_generated_context.nix) // args;
  batouModule = import {workdir}/{name}.nix moduleArgs;
in
{{ imports = [ batouModule ]; }}
"""


class NixOSModuleContext(Component):

    source_component: Component = None
    prefix: str = None

    def configure(self):

        if self.source_component:
            component = self.source_component
        else:
            component = self.parent

        context = nix_dict_to_nix({"component": component_to_nix(component)})

        if self.prefix is None:
            self.prefix = component.__class__.__name__.lower()

        self += NixFile(
            f"{self.prefix}_generated_context.nix",
            content=context,
            format_nix_code=True,
            sensitive_data=True,
        )


class NixOSModule(Component):

    namevar = "name"

    name: str
    path: Path = Path("/etc/local/nixos")
    context = None

    def configure(self):

        self += NixFile(f"{self.name}.nix")

        if self.context is None:
            if hasattr(self.parent, "nixos_context"):
                self.context = self.parent.nixos_context
            else:
                self.context = NixOSModuleContext(source_component=self.parent)
                self.parent.nixos_context = self.context
                self.parent += self.context

        self += NixFile(
            self.path / f"batou_{self.name}.nix",
            content=GLUE_MODULE_TEMPLATE.format(
                workdir=self.workdir, name=self.name, prefix=self.context.prefix
            ),
            format_nix_code=True,
        )


class PurgeNixOSModule(Purge):
    namevar = "name"

    name: str
    path: Path = Path("/etc/local/nixos")

    def configure(self):
        self.pattern = self.path / f"batou_{self.name}.nix"
        super().configure()



import json

import batou.component
import batou.lib.file
import batou.utils

import batou_ext.config
import batou_ext.nix


@batou_ext.nix.rebuild
class Memcached(batou.component.Component):
    """
    Helps to configure memcached

    Usage:
    self += batou_ext.memached.Memcached(
        port=11211,
        custom_config=dict(foo='baa', zzz=11))
    """

    port = batou.component.Attribute(int, 11211)

    # A additional dict
    custom_config = {}

    def configure(self):
        self.provide("memcached", self)
        self.address = batou.utils.Address(self.host.fqdn, self.port)

        # Trying to set some sane defaults
        base_config = dict(
            maxMemory=1024,
            port=self.port,
            maxConnections=1024,
        )

        config = batou_ext.config.dict_merge(base_config, self.custom_config)

        self += batou.lib.file.File(
            "/etc/local/memcached/memcached.json", content=json.dumps(config)
        )



import pkg_resources
from batou.component import Attribute, Component
from batou.lib.file import File


class ScalableService(Component):
    """
    Configure a systemd unit that can have multiple equal instances, such
    as multiple consumers for a message queue.

    Usage::

        self += ScalableService("message-consumer", running_instances=5)
        self.unit_identifier = self._.unit_identifier
        self += File("/etc/local/nixos/message-consumer-settings.nix")
        self += Rebuild()

    With a `message-consumer-settings.nix` looking like this:

        {
          systemd.services."{{ component.unit_identifier }}" = {
            serviceConfig = {
              /* ExecStart etc. */
            };
          };
        }

    This creates a systemd target called `message-consumer.target`
    and a template service called `message-consumer@.service`. When
    the configuration gets activated, five instances of this service
    are started, namely `message-consumer@0.service` to
    `message-consumer@4.service`. When the machine gets rebooted,
    these five instances will be started automatically.

    The number of instances running by default can be changed with the
    `running_instances` attribute.

    Please note that the unit is completely empty and thus invalid. A rebuild
    must not happen before `message-consumer-settings.nix` in the example above
    is added that fully configures the service.

    If needed, more services from this template can be started like this:

        $ systemctl start message-consumer@{5..23}

    This starts 19 additional units.

    When running

        $ systemctl restart message-consumer.target

    all 24 running units will be restarted.

    When running

        $ systemctl stop message-consumer.target

    all 24 running units will be stopped.

    However, when running

        $ systemctl start message-consumer.target

    after that, only 5 units will get back up. The same applies to an
    additional deployment or a reboot of the VM.

    Hence, starting additional services is a temporary measure to quickly
    get more workers up. The change must be persisted in the deployment
    after that.
    """

    namevar = "base_name"
    running_instances = Attribute(int, 4)

    def configure(self):
        self += File(
            f"/etc/local/nixos/scale-{self.base_name}.nix",
            content=pkg_resources.resource_string(
                "batou_ext", "resources/scalable-service.nix"
            ),
        )

        self.unit_identifier = f"{self.base_name}@"



"""Components for managing PHP.

Example for NixOS::


    class PHPApp(batou.component.Component):

        def configure(self):
            self.settings = self.require_one("settings")

            self += batou_ext.nix.UserEnv(
                "portal",
                packages=["php71", "php71Packages.redis"],
            )

            self += batou_ext.php.Ini(settings="", extensions=["redis.so"])
            self += batou_ext.php.FPM(
                "portal", dependency_strings=(self._.php_ini.content,)
            )

"""

import hashlib
import os.path

import batou.component
import batou.lib.file
import pkg_resources


class Ini(batou.component.Component):
    """Manage php.ini."""

    extensions = ()
    settings = ""
    logs = None

    def configure(self):
        self._extensions_dir = os.path.expanduser(
            "~/.nix-profile/lib/php/extensions/"
        )

        if self.logs is None:
            self.logs = self.map("logs")
        self += batou.lib.file.File(self.logs, leading=True, ensure="directory")
        self.error_log = os.path.join(self.logs, "php-error.log")
        self += batou.lib.logrotate.RotatedLogfile(
            self.error_log,
            postrotate="kill -USR1 $(cat {}/php-fpm.pid)".format(self.workdir),
        )

        # Providing a php.ini
        self += batou.lib.file.File(
            "php.ini",
            content=pkg_resources.resource_string(
                __name__, "resources/php/php.ini"
            ),
        )
        self.php_ini = self._


class FPM(batou.component.Component):
    """Provde running FPM.

    Usage::

        self += batou_ext.php.FPM("myphpproject")

    """

    name = None
    namevar = "name"

    php_ini = None  # defaults to `php.ini` in workdir
    logs = None  # defaults to `logs` in workdir

    global_settings = ""
    pool_settings = ""

    port = 9001

    env = {}  # Additional environmental values for FPM
    keep_env = batou.component.Attribute("literal", default=False)

    dependency_strings = ()

    def configure(self):
        self._checksum = hashlib.new("sha256")
        for s in self.dependency_strings:
            if isinstance(s, str):
                self._checksum.update(s.encode("utf-8"))
            else:
                self._checksum.update(s)

        self.address = batou.utils.Address(self.host.fqdn, self.port)

        # Logging
        if self.logs is None:
            self.logs = self.map("logs")
        self += batou.lib.file.File(self.logs, leading=True, ensure="directory")
        self.slow_log = os.path.join(self.logs, "slow.log")
        self += batou.lib.logrotate.RotatedLogfile(
            self.slow_log,
            postrotate="kill -USR1 $(cat {}/php-fpm.pid)".format(self.workdir),
        )

        # fpm.ini

        # Ensure we don't clear up environment if there is an env
        if self.env:
            self.keep_env = True

        self += batou.lib.file.File(
            "php-fpm.conf",
            content=pkg_resources.resource_string(
                __name__, "resources/php/php-fpm.conf"
            ),
        )
        self._checksum.update(self._.content)
        self.php_fpm_ini = self._.path

        if self.php_ini is None:
            self.php_ini = self.map("php.ini")

        # Start script
        self.pid_file = self.map("php-fpm.pid")
        self += batou.lib.file.File(
            self.name,
            mode=0o755,
            content=pkg_resources.resource_string(
                __name__, "resources/php/php-fpm.sh"
            ),
        )
        self._checksum.update(self._.content)

        # Service startup. We expect the service to be monitored by a generic
        # systemd "all proceses are running" check.
        self += batou.lib.service.Service(
            self.name,
            checksum=self._checksum.hexdigest(),
            systemd=dict(PIDFile=self.pid_file, Restart="always"),
        )



import os
import socket

from batou.component import Attribute, Component, ConfigString
from batou.lib.file import File
from batou.utils import Address


def resolve_v6(address):
    return socket.getaddrinfo(
        address.connect.host, address.connect.port, socket.AF_INET6
    )[0][4][0]


class PFAPostfix(Component):

    address = Attribute(Address, ConfigString("localhost:25"))

    def configure(self):
        self.address.listen.host_v6 = resolve_v6(self.address)

        self.db = self.require_one("pfa::database")
        self.keypair = self.require_one("keypair::mail")

        self.provide("postfix", self.address)

        self += File(
            "/etc/postfix/myhostname", content=self.address.connect.host
        )
        self += File(
            "/etc/postfix/main.d/40_local.cf", source=self.resource("local.cf")
        )
        self += File(
            "postfixadmin_virtual_alias",
            source=self.resource("postfixadmin_virtual_alias"),
        )
        self += File(
            "postfixadmin_virtual_domains",
            source=self.resource("postfixadmin_virtual_domains"),
        )
        self += File(
            "postfixadmin_virtual_sender_login",
            source=self.resource("postfixadmin_virtual_sender_login"),
        )
        self += File(
            "postfixadmin_virtual_mailboxes",
            source=self.resource("postfixadmin_virtual_mailboxes"),
        )

    def resource(self, filename):
        return os.path.join(os.path.dirname(__file__), "postfix", filename)
